import numpy as np
from scipy.interpolate import interp1d
from scipy import integrate
from scipy.stats import logistic
from .vasp_param import VaspParam
from .vasp_dos import VaspDOS
from .fermi_gas_dos import FermiGasDOS
from .fermi_gas_params import FGParams
from .constants import *
import warnings

warnings.filterwarnings("ignore")


def get_finite_Te_self_energy(outcar, algo, nkpts, Ef, core_E):
    """Function gets the energy and e-e scattering rates from the GW calculation
    Returns:
        E_gamma_occ (2D array - nkpts*nbands x 3): GW data of KS energies - Ef, scattering rate, and occupation values (0 to 1) sorted in ascending order of KS energies

    """

    ncol = 11
    head_string = "QP shifts evaluated in"
    tail_string = "Energies using frozen"
    header_num = [
        line_num for line_num, line in enumerate(outcar) if head_string in line
    ]
    tail_num = [line_num for line_num, line in enumerate(outcar) if tail_string in line]
    # For non-single shot calculations
    if algo.lower() == "gwr" or algo.lower() == "scgwr":
        tail_string_gwr = "ROTSIGMA"
        tail_num_gw = [
            line_num for line_num, line in enumerate(outcar) if tail_string_gwr in line
        ]
        tail_num = [tail_num[0]] + tail_num_gw

    def get_clean_data(data):
        return (
            np.array(
                [
                    [
                        x.split()
                        for x in [x.strip() for x in data if len(x.split()) == ncol]
                    ]
                ]
            )
            .flatten()
            .astype(float)
        )

    finite_data = [
        get_clean_data(outcar[head:tail])
        for head, tail in list(zip(header_num, tail_num))
    ]

    nbands = int(len(finite_data[0]) / (nkpts * ncol))

    KS_energies = np.array(
        [
            x[1::ncol].reshape(nkpts, nbands, int(len(x) / nkpts / nbands / ncol))
            for x in finite_data
        ]
    ).flatten()
    occ = (
        np.array(
            [
                x[7::ncol].reshape(nkpts, nbands, int(len(x) / nkpts / nbands / ncol))
                for x in finite_data
            ]
        ).flatten()
        / 2
    )
    gamma = (
        -2
        / hbar_eVfs
        * np.array(
            [
                x[8::ncol].reshape(nkpts, nbands, int(len(x) / nkpts / nbands / ncol))
                for x in finite_data
            ]
        ).flatten()
    )
    # sort in order of KS-energies
    E_gamma_occ = np.array(
        sorted(np.array([KS_energies - Ef, gamma, occ]).T, key=lambda x: x[0])
    )

    # get rid of any core states
    E_gamma_occ = np.array([x for x in E_gamma_occ if x[0] >= core_E])

    return E_gamma_occ[:, 0], E_gamma_occ[:, 1], E_gamma_occ[:, 2]


#!---------------------------------------------------------------
# * Start DOS and <gamma> section


def check_core_E_contribution(Te, core_E):
    core_E_occ = logistic.sf(core_E, 0, Te)
    if core_E_occ < 0.9999:
        print(
            f"\nFor Te = {Te:.2f} eV, states below the core_E of {core_E} contribute since the occupation of {core_E_occ} is less than 0.9999!\n"
        )


# Function gets the DOS predicted by VASP then interpolates for data matching
# The chemical potential should always be 0 eV for the scattering rates that are going to be used with this dos
def get_vasp_dos(Te, N, E, dos_file):
    # gets tet. method DOS
    if dos_file != None:
        data = VaspDOS(dos_file, Te, N)
        # mu is set to 0 eV
        interp_vasp_dos = interp1d(data.E - data.mu, data.dos)
        return interp_vasp_dos(E)
    else:
        print("\n\nNo VASP DOS .h5 file included in GWAvgGamma! Goodbye!\n\n")
        return [0] * len(E)


def avg_gamma_vasp_dos(Te, N, E, gamma, occ, dos_file, core_E):
    dos = get_vasp_dos(Te, N, E, dos_file)
    avg_gamma = integrate.simpson(gamma * occ * dos, E)
    check_core_E_contribution(Te, core_E)
    return avg_gamma


# get FG dos and then interpolates for data matching
# The chemical potential should always be 0 eV for the scattering rates that are going to be used with this dos
def get_fg_dos(Te, N, V, E):
    # need this so that there is no interpolation out of range issue when shifting by mu
    E_range = np.linspace(-500, 5000, 1000000)
    fg_params = FGParams(Te, N, V)
    fg_dos = FermiGasDOS(E_range, V)
    interp_fg_dos = interp1d(E_range - fg_params.mu, fg_dos.dos)
    # mu is set to 0 eV
    return interp_fg_dos(E)


def avg_gamma_fg_dos(Te, N, V, E, gamma, occ, core_E):
    dos = get_fg_dos(Te, N, V, E)
    avg_gamma = integrate.simpson(gamma * occ * dos, E)
    check_core_E_contribution(Te, core_E)
    return avg_gamma


# * End DOS and <gamma> section
#!---------------------------------------------------------------


class GWGamma:
    """
    class GWGamma generates all necessary parameters and outputs for analysis
    """

    def __init__(self, file, core_E, dos_file=None) -> None:
        """Function returns all parameters

        Args:
            file (string): the path to the OUTCAR file
            outcar (list): the OUTCAR file written to a list
            nkpts (int): the number of symmetry reduced KPOINTS
            algo (string): GW calculation algorithm
            Ef (float): Fermi energy of calculation, treated as the chemical potential
            core_E (float): cutoff (below) energy in eV that gets rid of core states - needed for integration
        """
        self.file = file
        self.outcar = [line for line in open(self.file, "r")]
        self.nkpts = int(
            [line for line in self.outcar if "NKPTS" in line][0].split()[3]
        )
        self.algo = [line for line in self.outcar if "ALGO" in line][0].split()[-1]
        self.Ef = float(
            [line for line in self.outcar if "E-fermi" in line][0].split()[-1]
        )
        self.core_E = core_E  # eV
        params = VaspParam(file)
        self.Te = params.Te  # eV
        self.N = params.N
        self.V = params.volume
        E_mu, gamma, occ = get_finite_Te_self_energy(
            self.outcar, self.algo, self.nkpts, self.Ef, self.core_E
        )
        self.E_mu = E_mu
        self.occ = occ
        self.gamma = gamma * (1 - occ)
        self.qh_gamma = gamma * occ
        self.total_gamma = gamma
        self.avg_vasp_gamma = avg_gamma_vasp_dos(
            params.Te, params.N, E_mu, gamma * (1 - occ), occ, dos_file, core_E
        )
        self.qh_avg_vasp_gamma = avg_gamma_vasp_dos(
            params.Te, params.N, E_mu, gamma * occ, (1 - occ), dos_file, core_E
        )
        self.avg_fg_gamma = avg_gamma_fg_dos(
            params.Te,
            params.N,
            params.volume,
            E_mu,
            gamma * (1 - occ),
            occ,
            dos_file,
            core_E,
        )
        self.qh_avg_fg_gamma = avg_gamma_fg_dos(
            params.Te,
            params.N,
            params.volume,
            E_mu,
            gamma * occ,
            (1 - occ),
            dos_file,
            core_E,
        )



import numpy as np
from scipy.stats import logistic
from scipy.interpolate import interp1d
from scipy import integrate
from .fermi_liquid_gamma import FLGamma
from .fermi_gas_dos import FermiGasDOS
from .fermi_gas_params import FGParams
import warnings

warnings.filterwarnings("ignore")


class FLAvgGamma:
    """
    class FLAvgGamma calculates the energy-averaged scattering rate predicted by the Fermil liquid following the Vignale method
    """

    def __init__(
        self,
        Te,
        N,
        V,
        E=np.linspace(-500, 500, 100000),
        occ_threshold=1e-8,
    ) -> None:
        """Function returns below parameters

        Args:
            Te (float): electronic temperature in eV
            N (int): Number of electrons in cell
            V (float): Volume of cell in Angstrom cubed
            E (array): Energy array in eV from -500 to 500 eV
            occ_threshold (float): threshold for highest energy to be considered "0" occupation
            E_qph (2D array): Energy-mu and total quasiparticle and quasihole scattering rates
            E_qh (2D array): Energy-mu and  quasihole scattering rates
            E_qp (2D array): Energy-mu and  quasiparticle scattering rates

        """
        self.E = E
        self.Te = Te
        self.N = int(N)
        self.V = V
        self.occ_threshold = occ_threshold

        fl_data = FLGamma(self.E, self.Te, self.N, self.V)
        gamma_data = fl_data.gamma()
        self.E_qph = gamma_data[0]
        self.E_qh = gamma_data[1]
        self.E_qp = gamma_data[2]

    # get FG dos and then interpolates for data matching
    # The chemical potential should always be 0 eV for the scattering rates that are going to be used with this dos
    def get_fg_dos(self):
        # need this so that there is no interpolation out of range issue when shifting by mu
        E_range = np.linspace(-500, 5000, 1000000)
        fg_params = FGParams(self.Te, self.N, self.V)
        fg_dos = FermiGasDOS(E_range, self.V)
        interp_fg_dos = interp1d(
            E_range - fg_params.mu,
            fg_dos.dos,
            fill_value=(0, np.nan),
            bounds_error=False,
        )
        # mu is set to 0 eV
        return interp_fg_dos(self.E_qph[:, 0])

    # FD distribution
    def get_occ(self):
        occ_condition = True  # turns to false if occ at highest energy does not meet certain threshold
        occ = logistic.sf(self.E_qph[:, 0], 0, self.Te)
        if occ[-1] > self.occ_threshold:
            print(
                f"occupation at {self.E_qph[:, 0][-1]} eV is {occ[-1]} for Te = {self.Te} eV"
            )
            occ_condition = False  # occ. not zero for highest energy
        return occ, occ_condition

    def fl_gamma_occ_dos(self):
        E = self.E_qp[:, 0]
        gamma = self.E_qp[:, 1]
        occ, occ_condition = FLAvgGamma.get_occ(self)
        dos = FLAvgGamma.get_fg_dos(self)
        return E, gamma, occ, occ_condition, dos

    def avg_fl_gamma(self):
        E, gamma, occ, occ_condition, dos = FLAvgGamma.fl_gamma_occ_dos(self)
        avg_gamma = integrate.simpson(gamma * occ * dos, E)
        return avg_gamma, occ_condition



import numpy as np
from scipy import integrate
from scipy.stats import logistic
import plasmapy.formulary as pp_form
import astropy.units as units
from .constants import *
import warnings
from .fermi_gas_dos import FermiGasDOS

warnings.filterwarnings("ignore")


#!---------------------------------------------------------------
# * Start of density- and theta-dependent parameter section


def get_kF(n):
    return (3 * np.pi**2 * n) ** (1 / 3)


def get_T_from_theta(n, theta):
    Ef = pp_form.quantum.Ef_(n * units.m**-3).value
    return theta * Ef / kB


def get_mu_quantum(n, theta):
    T = get_T_from_theta(n, theta)
    Ef = pp_form.quantum.Ef_(n * units.m**-3).value
    return Ef * (1 - 1 / 3 * ((np.pi * kB * T) / (2 * Ef)) ** 2)


# Eq. 5.26 on pp. 40 in Fetter and Walecka
def get_mu_classical(n, theta):
    T = get_T_from_theta(n, theta)
    mu = kB * T * np.log(n / 2 * pp_form.quantum.lambdaDB_th_(T * units.K).value ** 3)
    return mu


def get_mu(n, theta, theta_cutoff=1):
    if theta < theta_cutoff:
        mu = get_mu_quantum(n, theta)
    elif theta >= theta_cutoff:
        mu = get_mu_classical(n, theta)
    return mu


def get_omega_p(n):
    return pp_form.plasma_frequency(n * units.m**-3, particle="e-").value


# * End of density-dependent parameter section
#!---------------------------------------------------------------

#!---------------------------------------------------------------
# * Start getting the chemical potential by integrating over the occ*DOS section


# binary search function
def binary_search(Te, N, E, dos, bs_E):
    bs_N = [integrate.simpson(logistic.sf(E, Ef, Te) * dos, E) for Ef in bs_E]
    if np.round(bs_N[1], 6) == N:
        return bs_E[1], bs_N[1]
    elif bs_N[1] > N:
        return binary_search(Te, N, E, dos, [bs_E[0], np.mean(bs_E[:2]), bs_E[1]])
    elif bs_N[1] < N:
        return binary_search(Te, N, E, dos, [bs_E[1], np.mean(bs_E[1:]), bs_E[2]])


# integrated DOS*occ to find the chemical potential for given Te
def integrate_for_mu(Te, N, E, dos, search_min=-10000, search_max=10000):
    Ef, Ni = binary_search(Te, N, E, dos, [search_min, 0, search_max])
    return Ef, Ni


# * End getting the chemical potential by integrating over the occ*DOS section
#!---------------------------------------------------------------

#!---------------------------------------------------------------
# * Start of Eq. (4) section


def get_u_integral(u, k_kF, q_kF, theta, sq_mu):
    integral_first_term = (1 / (1 - np.exp(((2 * q_kF) / theta) * u))) * (
        1
        / (
            1
            + np.exp((-(2 * q_kF) / theta) * u)
            * np.exp((1 / theta) * (sq_mu - k_kF**2))
        )
    )
    integral_second_term = np.log(
        (1 + np.exp((1 / theta) * (sq_mu - (u + q_kF / 2) ** 2)))
        / (1 + np.exp((1 / theta) * (sq_mu - (-u + q_kF / 2) ** 2)))
    )

    return integral_first_term * integral_second_term


def get_G(k_kF, q_kF, theta, sq_mu, F_12, F_n12):

    integral, error = integrate.quad(
        get_u_integral,
        q_kF / 2 - k_kF,
        q_kF / 2 + k_kF,
        args=(k_kF, q_kF, theta, sq_mu),
    )

    G = (
        (9 * np.pi / (16 * np.sqrt(2)))
        * (theta ** (5 / 2))
        * (F_12 / F_n12) ** (3 / 2)
        * (q_kF / k_kF)
        * integral
    )

    return G, error


# * End of Eq. (4) section
#!---------------------------------------------------------------

#!---------------------------------------------------------------
# * Start of v(q) section


def get_k_sc(n, theta, F_12, F_n12):
    T = get_T_from_theta(n, theta)
    k_DH = 1 / pp_form.lengths.Debye_length(T * units.K, n * units.m**-3).value  # 1/m
    k_sc = k_DH * np.sqrt(F_n12 / F_12)  # 1/m

    return k_sc


def get_v(q, k_sc):
    v = 1 / (q**2 + k_sc**2)
    return v


# * End of v(q) section
#!---------------------------------------------------------------


#!---------------------------------------------------------------
# * Start of Eq. (3) section


def get_q_integrand(q, q_kF, k_kF, k_sc, theta, sq_mu, F_12, F_n12):
    v2 = np.abs(get_v(q, k_sc)) ** 2
    G = get_G(k_kF, q_kF, theta, sq_mu, F_12, F_n12)[0]

    q_integrand = v2 * G
    return q_integrand


def get_Eq_3(q_list, kF, k_kF, k_sc, theta, sq_mu, F_12, F_n12):

    q_integrand = np.array(
        [
            get_q_integrand(q, q / kF, k_kF, k_sc, theta, sq_mu, F_12, F_n12)
            for q in q_list
        ]
    )
    q_integral = integrate.cumtrapz(q_integrand, q_list, initial=0)[-1]

    return k_sc**3 * q_integral


def get_gamma_omega_p(q_kF_list, k_kF_list, n, theta, mu):
    # * Assign values independent of k and q
    kF = get_kF(n)
    sq_mu = mu / pp_form.quantum.Ef_(n * units.m**-3).value
    F_12 = np.real(pp_form.mathematics.Fermi_integral(sq_mu / theta, 1 / 2))
    F_n12 = np.real(pp_form.mathematics.Fermi_integral(sq_mu / theta, -1 / 2))
    k_sc = get_k_sc(n, theta, F_12, F_n12)
    q_list = q_kF_list * get_kF(n)
    k_list = k_kF_list * get_kF(n)

    gamma_omega_p = np.array(
        [get_Eq_3(q_list, kF, k / kF, k_sc, theta, sq_mu, F_12, F_n12) for k in k_list]
    )

    return gamma_omega_p


# * End of Eq. (3) section
#!---------------------------------------------------------------

#!---------------------------------------------------------------
# * Start of Gamma section


def get_gamma(E, n, mu, theta, q_kF_list):
    """
    Function calculates Daligault gamma
    Returns:
        E_mu (1D array): Energy-mu (eV)
        gamma (1D array):  Daligault scattering rate (1/fs)
        dos (1D array): Fermi gas DOS
        occ (1D array): Fermi dirac occupation values
    """

    k_kF_list = (np.sqrt(2 * m_e * E / J_to_eV) / hbar_Js) / get_kF(n)

    gamma = (
        np.array(get_gamma_omega_p(q_kF_list, k_kF_list, n, theta, mu / J_to_eV))
        * get_omega_p(n)
        * 1e-15
    )
    # 1/fs

    return gamma


def get_avg_gamma(gamma, dos, occ, E_mu):
    return integrate.simpson(gamma * dos * occ, E_mu)


# * End of Gamma section
#!---------------------------------------------------------------


class DaligaultGamma:
    """
    class DaligaultGamma generates all necessary parameters and outputs for analysis
    """

    def __init__(
        self, E, Te, N, V, theta_cutoff=1, q_kF_list=np.linspace(1e-3, 10, 1000)
    ) -> None:
        """Function returns all parameters

        Args:
            E (array): Energy array in eV
            Te (float): electronic temperature in eV
            N (int): Number of electrons in cell
            V (float): Volume of cell in Angstrom cubed (A^3)
            n (float): carrier density of material (1/m^3)
            theta (float): degeneracy parameter
            theta_cutoff (float): Te (eV) cutoff in which we go from quantum to classical chemical potential
            q_kF_list (array): array of q/kF points used for v(q) and G(q/kF), defaults to np.linspace(1e-3, 10, 1000) 1/m
            model_mu (float): Chemical potential in eV obtained from either a FG or classical model depending on theta
            dos (array): Array of Fermi gas density of states
            mu (float): Chemical potential in eV obtained by integrating the DOS*occupation over energy and constraining by N
            occ (array): Fermi Dirac occupation values, where the chemical potential is set to 0 eV
            gamma (array): Daligualt method of estimating gamma(E)
            avg_gamma (float): gamma*occ*dos integrated over energy
            E_mu (array): Energy array minus mu in eV
        """
        self.E = E
        self.Te = Te
        self.N = N
        self.V = V
        self.n = N / (V * angstrom**3)
        self.theta = pp_form.quantum_theta(self.Te * units.eV, self.n).value
        self.theta_cutoff = theta_cutoff
        self.q_kF_list = q_kF_list
        self.model_mu = get_mu(self.n, self.theta, self.theta_cutoff) * J_to_eV
        fg_dos = FermiGasDOS(self.E, self.V)
        self.dos = fg_dos.dos
        self.mu = integrate_for_mu(self.Te, self.N, self.E, self.dos)[0]
        self.occ = logistic.sf(self.E, self.mu, self.Te)
        self.gamma = get_gamma(self.E, self.n, self.mu, self.theta, self.q_kF_list)
        self.avg_gamma = get_avg_gamma(self.gamma, self.dos, self.occ, self.E)
        self.E_mu = self.E - self.mu



import numpy as np
from scipy import integrate
from scipy.stats import logistic
from scipy.interpolate import interp1d
import math
import warnings
from .constants import *
from .fermi_gas_params import FGParams
from .fermi_gas_dos import FermiGasDOS

warnings.filterwarnings("ignore")


#!---------------------------------------------------------------
# * Start getting the chemical potential by integrating over the occ*DOS section


# binary search function
def binary_search(Te, N, E, dos, bs_E):
    bs_N = [integrate.simpson(logistic.sf(E, Ef, Te) * dos, E) for Ef in bs_E]
    if np.round(bs_N[1], 6) == N:
        return bs_E[1], bs_N[1]
    elif bs_N[1] > N:
        return binary_search(Te, N, E, dos, [bs_E[0], np.mean(bs_E[:2]), bs_E[1]])
    elif bs_N[1] < N:
        return binary_search(Te, N, E, dos, [bs_E[1], np.mean(bs_E[1:]), bs_E[2]])


# integrated DOS*occ to find the chemical potential for given Te
def integrate_for_mu(Te, N, E, dos, search_min=-10000, search_max=10000):
    Ef, Ni = binary_search(Te, N, E, dos, [search_min, 0, search_max])
    return Ef, Ni


# * End getting the chemical potential by integrating over the occ*DOS section
#!---------------------------------------------------------------

#!---------------------------------------------------------------
# * Start gamma sections


def get_qh_gamma(E, mu, Te, Ef, xi3):
    qh_gamma = (
        np.pi
        / (8 * hbar_eVfs * Ef)
        * ((mu - E) ** 2 + (np.pi * Te) ** 2)
        / (1 + np.exp(-(mu - E) / Te))
        * xi3
    )
    return qh_gamma


def get_qp_gamma(E, mu, Te, Ef, xi3):
    qp_gamma = (
        np.pi
        / (8 * hbar_eVfs * Ef)
        * ((E - mu) ** 2 + (np.pi * Te) ** 2)
        / (1 + np.exp(-(E - mu) / Te))
        * xi3
    )
    return qp_gamma


def get_avg_gamma(gamma, dos, occ, E):
    return integrate.simpson(gamma * dos * occ, E)


# * End gamma sections
#!---------------------------------------------------------------


class VignaleGamma:
    """
    class VignaleGamma generates all necessary parameters and outputs for analysis
    """

    def __init__(self, E, Te, N, V) -> None:
        """Function returns all parameters

        Args:
            E (array): Energy array in eV
            Te (float): electronic temperature in eV
            N (int): Number of electrons in cell
            V (float): Volume of cell in Angstrom cubed
            Ef (float): Fermi energy of free electron gas in eV
            mu (float): Chemical potential of free electron gas in eV
            V_au (float): Volume in atomic units
            n_au (float): particle density (N/V) in a.u.
            kf (float): Fermi wave vector in a.u.
            rs (float): Wigner-Seitz radius in a.u.
            N0 (float): DOS at Ef in a.u.
            a3 (float): alpha parameters in a.u.
            xi3 (float): xi3 parameters in a.u.
        """
        self.E = E
        self.Te = Te
        self.N = int(N)
        self.V = V
        fg_params = FGParams(self.Te, self.N, self.V)
        self.Ef = fg_params.Ef
        self.model_mu = fg_params.mu
        self.V_au = fg_params.V_au
        self.kf = fg_params.kf
        self.rs = fg_params.rs
        self.N0 = fg_params.N0
        self.a3 = np.pi**2 * self.N0 / (self.kf**2 * self.rs)
        self.xi3 = np.sqrt((self.a3 * self.rs) / (4 * np.pi)) * math.atan(
            np.sqrt(np.pi / (self.a3 * self.rs))
        ) + 1 / (2 * (1 + np.pi / (self.a3 * self.rs)))
        #! Here, think about how to handle below
        fg_dos = FermiGasDOS(self.E, self.V)
        self.dos = fg_dos.dos
        self.mu = integrate_for_mu(self.Te, self.N, self.E, self.dos)[0]
        self.occ = logistic.sf(self.E, self.mu, self.Te)
        self.qh_gamma = get_qh_gamma(self.E, self.mu, self.Te, self.Ef, self.xi3)
        self.qp_gamma = get_qp_gamma(self.E, self.mu, self.Te, self.Ef, self.xi3)
        self.total_gamma = self.qh_gamma + self.qp_gamma
        self.avg_gamma = get_avg_gamma(self.qp_gamma, self.dos, self.occ, self.E)
        self.E_mu = self.E - self.mu



import numpy as np
import math
import warnings
from .constants import *

warnings.filterwarnings("ignore")


class FLGamma:
    """
    class FLGamma generates all necessary parameters and outputs for analysis
    """

    def __init__(self, E, Te, N, V) -> None:
        """Function returns all parameters

        Args:
            E (array): Energy array in eV
            Te (float): electronic temperature in eV
            N (int): Number of electrons in cell
            V (float): Volume of cell in Angstrom cubed
            Ef (float): Fermi energy of free electron gas in eV
            mu (float): Chemical potential of free electron gas in eV
            V_au (float): Volume in atomic units
            n_au (float): particle density (N/V) in a.u.
            kf (float): Fermi wave vector in a.u.
            rs (float): Wigner-Seitz radius in a.u.
            N0 (float): DOS at Ef in a.u.
            a3 (float): alpha parameters in a.u.
            xi3 (float): xi3 parameters in a.u.
        """
        self.E = E
        self.Te = Te
        self.N = int(N)
        self.V = V
        self.Ef = (
            hbar_Js**2
            / (2 * m_e)
            * (3 * np.pi**2 * self.N / (self.V * 1e-30)) ** (2 / 3)
        ) * J_to_eV
        self.mu = self.Ef * (1 - 1 / 3 * ((np.pi * self.Te) / (2 * self.Ef)) ** 2)
        # Use atomic units:  hbar = m = e = 1  a.u
        self.V_au = self.V * (1.88973) ** 3  # a0^3
        self.n_au = self.N / self.V_au
        self.kf = (3 * np.pi**2 * self.n_au) ** (1 / 3)
        self.rs = (3 / (4 * np.pi * self.n_au)) ** (1 / 3)
        self.N0 = 3 * self.n_au / (self.kf**2)
        self.a3 = np.pi**2 * self.N0 / (self.kf**2 * self.rs)
        self.xi3 = np.sqrt((self.a3 * self.rs) / (4 * np.pi)) * math.atan(
            np.sqrt(np.pi / (self.a3 * self.rs))
        ) + 1 / (2 * (1 + np.pi / (self.a3 * self.rs)))

    def gamma(self):
        """
        Function calculates Vignale FL gamma
        Returns:
            E_qph (2D array): Energy-mu and total quasiparticle and quasihole scattering rates
            E_qh (2D array): Energy-mu and  quasihole scattering rates
            E_qp (2D array): Energy-mu and  quasiparticle scattering rates
        """
        qh_gamma = (
            np.pi
            / (8 * hbar_eVfs * self.Ef)
            * ((self.mu - self.E) ** 2 + (np.pi * self.Te) ** 2)
            / (1 + np.exp(-(self.mu - self.E) / self.Te))
            * self.xi3
        )
        qp_gamma = (
            np.pi
            / (8 * hbar_eVfs * self.Ef)
            * ((self.E - self.mu) ** 2 + (np.pi * self.Te) ** 2)
            / (1 + np.exp(-(self.E - self.mu) / self.Te))
            * self.xi3
        )

        E_qph = np.array([self.E - self.mu, qh_gamma + qp_gamma]).T
        E_qh = np.array([self.E - self.mu, qh_gamma]).T
        E_qp = np.array([self.E - self.mu, qp_gamma]).T
        return E_qph, E_qh, E_qp



import numpy as np
import warnings
from .constants import *

warnings.filterwarnings("ignore")


class FermiGasDOS:
    """
    class FermiGasDOS calculates and returns the DOS predicted by the Fermi gas model
    """

    def __init__(self, E, V) -> None:
        """Function returns all E and DOS

        Args:
            E (array): Energy array in eV
            V (float): Volume of cell in Angstrom cubed
            dos (array): Density of states predicted by FG and all nan is turned into 0

        """
        self.E = E
        self.V = V
        self.dos = np.nan_to_num(
            np.array(
                (
                    ((self.V * 1e-30) / (2 * np.pi**2))
                    * (2 * m_e / (hbar_Js**2)) ** (3 / 2)
                )
                / J_to_eV ** (3 / 2)
                * np.sqrt(self.E)
            )
        )



import numpy as np
from scipy.interpolate import interp1d
from scipy import integrate
from scipy.stats import logistic
from .gw_gamma import GWGamma
from .vasp_param import VaspParam
from .vasp_dos import VaspDOS
from .fermi_gas_dos import FermiGasDOS
from .fermi_gas_params import FGParams
import warnings

warnings.filterwarnings("ignore")

# TODO: add in warning when temperatures lead to core energy contributions!


class GWAvgGamma:
    """
    class GWAvgGamma calculates the energy-averaged scattering rate
    """

    def __init__(self, file, core_E, dos_file=None) -> None:
        """Function returns below parameters

        Args:
            file (string): directory and name of VASP OUTCAR file
            dos_file (string, optional): directory and name of VASP DOS .h5 file
            core_E (float): cutoff (below) energy in eV that gets rid of core states - needed for integration
            E_gamma_occ (2D array - nkpts*nbands x 3): GW data of KS energies - Ef, scattering rate, and occupation values (0 to 1) sorted in ascending order of KS energies
            Te (float): electronic temperature in eV
            V (float): Volume of cell in A^3
            N (int): Number of electrons accounting for degeneracy

        """
        self.file = file
        self.dos_file = dos_file
        self.core_E = core_E  # eV
        params = VaspParam(file)
        gw_data = GWGamma(file, self.core_E)
        self.E_gamma_occ = gw_data.get_finite_Te_self_energy()
        self.Te = params.Te  # eV
        self.N = params.N
        self.V = params.volume

    # Function checks to ensure that the occupation of the core energy does not contribute to thermal excitations
    def check_core_E_contribution(self):
        core_E_occ = logistic.sf(self.core_E, 0, self.Te)
        if core_E_occ < 0.9999:
            print(
                f"States below the core_E of {self.core_E} contribute since the occupation of {core_E_occ} is less than 0.9999!"
            )

    # Function gets the DOS predicted by VASP then interpolates for data matching
    # The chemical potential should always be 0 eV for the scattering rates that are going to be used with this dos
    def get_vasp_dos(self):
        # gets tet. method DOS
        if self.dos_file != None:
            data = VaspDOS(self.dos_file, self.Te, self.N)
            # mu is set to 0 eV
            interp_vasp_dos = interp1d(data.E - data.mu, data.dos)
            return interp_vasp_dos(self.E_gamma_occ[:, 0])
        else:
            print("\n\nNo VASP DOS .h5 file included in GWAvgGamma! Goodbye!\n\n")
            exit()

    def gw_gamma_occ_vasp_dos(self):
        E = self.E_gamma_occ[:, 0]
        gamma = self.E_gamma_occ[:, 1]
        occ = self.E_gamma_occ[:, 2]
        dos = GWAvgGamma.get_vasp_dos(self)
        return E, gamma, occ, dos

    def avg_gw_gamma_vasp_dos(self):
        E, gamma, occ, dos = GWAvgGamma.gw_gamma_occ_vasp_dos(self)
        avg_gamma = integrate.simpson(gamma * occ * (1 - occ) * dos, E)
        GWAvgGamma.check_core_E_contribution(self)
        return avg_gamma

    # get FG dos and then interpolates for data matching
    # The chemical potential should always be 0 eV for the scattering rates that are going to be used with this dos
    def get_fg_dos(self):
        # need this so that there is no interpolation out of range issue when shifting by mu
        E_range = np.linspace(-500, 5000, 1000000)
        fg_params = FGParams(self.Te, self.N, self.V)
        fg_dos = FermiGasDOS(E_range, self.V)
        interp_fg_dos = interp1d(E_range - fg_params.mu, fg_dos.dos)
        # mu is set to 0 eV
        return interp_fg_dos(self.E_gamma_occ[:, 0])

    def gw_gamma_occ_fg_dos(self):
        E = self.E_gamma_occ[:, 0]
        gamma = self.E_gamma_occ[:, 1]
        occ = self.E_gamma_occ[:, 2]
        dos = GWAvgGamma.get_fg_dos(self)
        return E, gamma, occ, dos

    def avg_gw_gamma_fg_dos(self):
        E, gamma, occ, dos = GWAvgGamma.gw_gamma_occ_fg_dos(self)
        avg_gamma = integrate.simpson(gamma * occ * (1 - occ) * dos, E)
        GWAvgGamma.check_core_E_contribution(self)
        return avg_gamma
